"""Core programmatic models for hpcflow.

EAR abort exit code is set to 64 [1].

References
----------
https://tldp.org/LDP/abs/html/exitcodes.html

"""

ALL_TEMPLATE_FORMATS = ("yaml", "json")
ABORT_EXIT_CODE = 64
